use anyhow::{Context, Result, bail};
use std::fs;

pub mod cli;
pub mod config;
pub mod downloader;

use cli::Args;
use config::Config;

/// The main application logic.
pub async fn run(args: Args) -> Result<()> {
    println!("▶️  Starting ruleset processor...");
    println!("    Config file:      {}", args.input.display());
    println!("    Output directory:   {}", args.output.display());
    println!("    Concurrency level:  {}", args.concurrency);

    // Load and parse the configuration file into strongly-typed structs.
    let config = Config::load(&args.input)
        .with_context(|| format!("Failed to load config from {}", args.input.display()))?;

    // Ensure the output directory exists before any operations.
    fs::create_dir_all(&args.output).with_context(|| {
        format!(
            "Failed to create output directory '{}'",
            args.output.display()
        )
    })?;

    // Determine the set of files that should exist based on the config.
    let expected_files = config
        .get_expected_files(&args.output)
        .context("Failed to determine expected files from config")?;

    // Clean up any files in the output directory that are not in our expected set.
    downloader::cleanup_stale_files(&args.output, &expected_files)?;

    // Proceed to download files defined in the config.
    let urls_to_download = config.extract_urls();
    if urls_to_download.is_empty() {
        println!("\n✔️  No rule sets with URLs found. Process complete.");
        return Ok(());
    }
    println!(
        "\n✔️  Found {} rule sets to download/update.",
        urls_to_download.len()
    );

    // Download all files concurrently.
    let download_report =
        downloader::download_all_rules(&urls_to_download, &args.output, args.concurrency).await?;

    println!("\n✅ Download process finished.");
    println!(
        "   {} successful, {} failed.",
        download_report.successful, download_report.failed
    );

    // If any downloads failed, abort with an error message.
    if download_report.failed > 0 {
        bail!(
            "Aborting due to {} download failures.",
            download_report.failed
        );
    }

    println!("\n✔️  Ruleset synchronization complete.");
    Ok(())
}
